# -- The settings table; has support for ttl etc. ---------------------------

require "json"

class MicroSql::KeyValueTable < MicroSql::Table
  def initialize(db, table_name = "settings")
    sql = if db.is_a?(MicroSql::PgAdapter)
      "CREATE TABLE #{table_name}_b(uid TEXT PRIMARY KEY, value BYTEA, ttl INTEGER NOT NULL)"
    else
      "CREATE TABLE #{table_name}(uid TEXT PRIMARY KEY, value TEXT, ttl INTEGER NOT NULL)"
    end

    super db, sql
  end
  
  def [](key)
    value, ttl = db.ask("SELECT value, ttl FROM #{table_name} WHERE uid=?", key)
    decode(value) if ttl && (ttl == 0 || ttl > Time.now.to_i)
  end

  def delete_all
    @db.exec "DELETE FROM #{table_name}"
  end

  def cached(key, ttl = nil, &block)
    self[key] || update(key, yield, ttl)
  end

  def update(key, value, ttl = nil)
    if value
      encoded = encode(value)
      ttl = ttl ? ttl + Time.now.to_i : 0
      affected = @db.ask("UPDATE #{table_name} SET value=?, ttl=? WHERE uid=?", encoded, ttl, key)
      if affected == 0
        @db.ask("INSERT INTO #{table_name}(value, ttl, uid) VALUES(?,?,?)", encoded, ttl, key)
      end
    else
      @db.ask("DELETE FROM #{table_name} WHERE uid=?", key)
    end

    value
  end

  def expire(key, ttl)
    ttl = ttl ? ttl + Time.now.to_i : ttl
    @db.ask("UPDATE #{table_name} SET ttl=? WHERE uid=?", ttl, key)
  end

  alias :[]= :update

  private
  
  def decode(io)
    return unless io
    return PG::Connection.unescape_bytea(io) if db.is_a?(MicroSql::PgAdapter)

    JSON.parse("[#{io}]").first
  end
  
  def encode(value)
    return unless value
    
    return db.connection.escape_bytea(value.to_s) if db.is_a?(MicroSql::PgAdapter)
    
    value.to_json
  end
end
